
package require struct::list

# This class implements inter-bot communication. It should be subclassed to
# implement transports e.g. IBCIRC for IRC transport in plugin.py Transport is
# implemented by two methods. do_send(self, to, *args) should be implemented in
# subclasses to send its arguments to another bot.  received(self, source,
# msgid, cmd, *args) should be called when a message for this bot is received.
# msgid and cmd are just a part of every message (included in do_send's args)

# methods in the API section are to notify other bots about this bot's events,
# request some data etc

# There are also a few callbacks (abstract methods) which should be implemented
# by subcasses and basically answer to another bot's API call.

::hlirc::define-module ibc {

	namespace path [list ::tcl::mathop ::tcl::mathfunc {*}[namespace path]]

	module require state

	module config bootstrap "" {Nicknames of bots used to bootstrap IBC facility from.}
	module config min-connections 5 {The bot will try to maintain more than this number of connections}
	module config max-connections 5 {The bot will try to maintain less than this number of connections}

	module wizard {
		puts "This bot has a few features for communication between
channels like delegates or plenaries. For it to work the
bots need to know each others' nicknames. Please enter
the list of CNT bots you know exist on your IRC network."
		set bootstrap [::hlirc::wizard query "Space-separated list of bot nicknames:" "" ::cnt::nicks-list? "Invalid nickname in the list."]
		module config bootstrap $bootstrap
	}

	variable connections {}
	variable pendingConnections
	variable listings
	variable listingsmeta
	variable listingscache
	array set listingscache {}

	variable timeout
	array set timeout {
		handshake    30
		listing      10
		listingcache 600
		cachesweep   60
	}

	proc timeout {type} {
		variable timeout
		return [* $timeout($type) 1000]
	}

	proc do-connected {} {
		module log info "Making initial connections..."
		struct::list map [split [module config bootstrap]] connect
		# Wait 30 seconds for bootstrap
		module after 30000 start-normalize-connections
		module after [timeout cachesweep] check-stale-listings
	}

	# Sent a message to a bot which is now offline
	proc do-err-nosuchnick {} {
		disconnect [additional 0]
	}

	# Commands

	variable handlers
	proc cmd-ibc-add-handler {command handler} {
		variable handlers
		set handlers($command) $handler
	}

	proc handle {command args} {
		variable handlers
		if {$command in [array names handlers]} {
			try {
				{*}$handlers($command) {*}$args
			} trap {TCL WRONGARGS} msg {
				module log critical "Calling $command with $args"
				error $msg $::errorInfo {TCL WRONGARGS}
			}
		} else {
			error "Unknown IBC command: $command. Should be one of: [array names handlers]"
		}
	}

	# Implementation

	proc check-connections {} {
		variable connections
		if {[llength $connections] < [module config min-connections] ||
			[llength $connections] > [module config max-connections]} {
			start-normalize-connections once
		}
	}

	command ib {} {
		variable pendingConnections
		variable connections
		variable listings
		variable listingsmeta
		variable handlers
		set source [who]
		match {msgid command {rest rest}} {
			if {![check-msgid $msgid]} {
				module log error "Malformed msgid"
				return
			}
			switch $command {
				# Connections
				syn {
					module log info "Connection attempt from $source"
					set pendingConnections($msgid) $source
					stale-item handshake pendingConnections $msgid
					send $source $msgid syn-ack
				}
				syn-ack {
					if {$msgid ni [array names pendingConnections]} {
						module log warn "Unexpected syn-ack. syn is either stale or wasn't received at all"
					} elseif {$source != $pendingConnections($msgid)} {
						module log warn "Connection with msgid $msgid. Expected from $pendingConnections($msgid), received from $source"
					} else {
						on-connected $source
						array unset pendingConnections $msgid
						struct::set include connections $source
						send $source $msgid ack
					}
					check-connections
				}
				ack {
					if {$msgid ni [array names pendingConnections]} {
						module log warn "Unexpected ack. syn-ack is either stale or wasn't received at all"
					} elseif {$source != $pendingConnections($msgid)} {
						module log warn "Connection with msgid $msgid. Expected from $pendingConnections($msgid), received from $source"
					} else {
						on-connected $source
						array unset pendingConnections $msgid
						struct::set include connections $source
					}
					check-connections
				}
				rst {
					module log info "Connection dropped with $source"
					cnt::lremove connections $source
					check-connections
				}
				# listings
				list {
					if {$msgid in [array names listings]} return
					set listings($msgid) true
					stale-item listing listings $msgid
					lassign $rest what client args
					set data [getdata $what {*}$args]
					if {$data != ""} {
						send $client $msgid listreport {*}$data
					}
					foreach connection $connections {
						if {$connection ni [list $source $client]} {
							send $connection $msgid $command {*}$rest
						}
					}
				}
				listreport {
					if {$msgid in [array names listingsmeta]} {
						lassign $listingsmeta($msgid) request user
						listing-data $user $rest
						lappend listings($msgid) $rest
					} else {
						module log warn "Received listreport for nonexistent message $msgid"
					}
				}
				error {
					module log error "Got error from $source: $rest"
				}
				default {
					handle $command {*}$rest
				}
			}
		}
	}

	proc on-connected {with} {
		variable listingscache
		module log info "Connection established with $with"
		if {[array names listingscache -exact bots] != ""} {
			array unset listingscache bots
		}
	}

	proc send {to msgid args} {
		if {$msgid == ""} {
			set msgid [makemsgid]
		}
		[irc] privmsg $to !ib $msgid $args
	}

	proc getdata {what args} {
		variable connections
		set nick [[irc] state nick]
		switch $what {
			connections {
				list $nick [llength $connections]
			}
			bots {
				set mainchannel [[irc] mainchannel]
				set allchannels [[irc] state channels]
				set channels [list $mainchannel {*}[struct::set difference $allchannels $mainchannel]]
				list $nick $channels
			}
			default {
				handle get-data $what {*}$args
			}
		}
	}

	proc listing-finished {msgid request user {coroutine ""}} {
		variable listings
		variable listingscache
		if {![info exists listings($msgid)]} {
			set listings($msgid) ""
		}
		set listingscache($request) [list [clock seconds] $listings($msgid)]
		if {$coroutine != ""} {
			{*}$coroutine
		}
		array unset listings $msgid
		array unset listingsmeta $msgid
	}

	proc listing-data {user data} {
		if {$user != ""} {
			[irc] privmsg $user [join $data]
		}
	}


	# Remove old cache entries
	proc check-stale-listings {} {
		variable listingscache
		set time [clock seconds]
		foreach what [array names listingscache] {
			lassign $listingscache($what) cachetime data
			if {$time > $cachetime + [timeout listingcache]} {
				array unset listingscache $what
			}
		}
		after [timeout cachesweep] [namespace current]::check-stale-listings
	}

	proc cachedlisting {name} {
		variable listingscache
		lindex $listingscache($name) 1
	}

	proc stale-item {timeouttype dict key} {
		after [timeout $timeouttype] [namespace current]::do-stale-item $dict $key
	}

	proc do-stale-item {dict key} {
		variable $dict
		if {$key in [array names dict]} {
			if {$dict == "pendingConnections"} {
				module log warn "Unused pendingConnections item: $key to [set ${dict}(${key})]"
			}
			array unset $dict $key
		}
	}

	proc normalize-connections {data} {
		variable connections
		# Filter out myself
		set data [struct::list filterfor item $data {
			[lindex $item 0] != [[irc] state nick]
		}]
		if {[llength $connections] < [module config min-connections]} {
			# Filter out bots we are already connected with
			set data [struct::list filterfor item $data {
				[lindex $item 0] ni $connections
			}]
			# Connect to bots with lowest connections
			set data [lsort -increasing -index 1 $data]
			set newconnections [min [llength $data] [- [module config min-connections] [llength $connections]]]
			for {set i 0} {$i < $newconnections} {incr i} {
				connect [lindex $data 0]
			}
		} elseif {[llength $connections] > [module config max-connections]} {
			# Filter out bots we are not connected with
			set data [struct::list filterfor item $data {
				[lindex $item 0] in $connections
			}]
			set data [lsort -decreasing -index 1 $data]
			set disconnectcount [- [llength $connections] [module config max-connections]]
			for {set i 0} {$i < $disconnectcount} {incr $i} {
				disconnect [lindex $data 0]
			}
		}
	}

	variable noConnectionsWarned false
	proc start-normalize-connections {{once ""}} {
		variable connections
		if {[llength $connections] == 0} {
			variable noConnectionsWarned
			if {!$noConnectionsWarned} {
				set noConnectionsWarned true
				module log warn "The bot is running without connections to other bots. Consider setting 'bootstrap' option in the config"
			}
		} elseif {[< [module config min-connections] [llength $connections] [module config max-connections]]} {
			# All is well
			return
		} else {
			module log info "Have [llength $connections] connections. Trying to normalize"
			listing "" connections [info coroutine]
			yield-irc [+ 1 [timeout listing]]
			normalize-connections [cachedlisting connections]
		}
		if {$once == ""} {
			module after [* 10 60 1000] start-normalize-connections
		}
	}

	# API

	proc connect {to} {
		variable pendingConnections
		module log info "Establishing connection to $to"
		set msgid [makemsgid]
		set pendingConnections($msgid) $to
		stale-item handshake pendingConnections $msgid
		send $to $msgid syn
	}

	proc disconnect {to} {
		variable connections
		# Because it can happen during handshake
		if {$to in $connections} {
			module log info "Dropping connection to $to"
			send $to "" rst
			cnt::lremove connections $to
		}
	}

	proc listing {user request {coroutine ""}} {
		variable listings
		variable listingsmeta
		variable listingscache
		variable connections
		set msgid [makemsgid]
		lassign $request what args
		# Check cache
		if {$request in [array names listingscache]} {
			foreach line [lindex $listingscache($request) 1] {
				listing-data $user $line
			}
			if {$coroutine != ""} {
				after idle $coroutine
			}
			return
		}

		# Retrieve data
		set data [getdata $what {*}$args]
		if {$data != ""} {
			set listings($msgid) [list $data]
			listing-data $user $data
		}
		set listingsmeta($msgid) [list $request $user]
		foreach connection $connections {
			send $connection $msgid list $what [[irc] state nick] {*}$args
		}
		module after [timeout listing] listing-finished $msgid $request $user $coroutine
	}

	proc cmd-ibc-listing {request {user ""}} {
		listing $user $request [info coroutine]
		yield-irc [+ 1000 [timeout listing]]
		cachedlisting $request
	}

	# Suspends current evaluation until this bot receives a message from channel's bot
	proc find-bot {channel} {
		# Check if we host the channel
		if {$channel == [[irc] mainchannel]} {
			return [[irc] state nick]
		}
		if {[[irc] db select-plenary $channel] != ""} {
			return [[irc] state nick]
		}
		try {
			# Try to get a list of bots on the network
			return [do-find-bot $channel]
		} trap {bot notfound} msg {
			# Flush the cache and try again (FIXME)
			cmd-ibc-drop-bots-cache
			try {
				return [do-find-bot $channel]
			} trap {bot notfound} msg {
				module log warn "Can't find the bot serving $channel"
				return
			}
		}
	}

	proc do-find-bot {channel} {
		set bots [all-bots]
		foreach bot $bots {
			set channels [struct::list map [lindex $bot 1] hlirc::canonicalize]
			if {[hlirc::canonicalize $channel] in $channels} {
				return [lindex $bot 0]
			}
		}
		throw {bot notfound} "find-bot was called for a channel which either doesn't have a bot or doesn't exist itself"
	}

	proc all-bots {} {
		tailcall cmd-ibc-listing bots
	}

	proc cmd-ibc-find-bot {channel} {
		tailcall find-bot $channel
	}

	proc cmd-ibc-drop-bots-cache {} {
		variable listingscache
		try {
			unset listingscache(bots)
		} trap {TCL LOOKUP ELEMENT} msg {
			return
		}
	}

	proc cmd-ibc-send {command channel args} {
		set bot [find-bot $channel]
		send $bot "" $command $channel {*}$args
	}

	# Util

	# Message ID is a string of unix time and random chars
	proc makemsgid {} {
		set timestamp [clock seconds]
		set string [cnt::random-string 6]
		return "$timestamp$string"
	}

	# Checks if a msgid is valid
	proc check-msgid {msgid} {
		set date [string range $msgid 0 9]
		# It should begin with a date which is not more than an hour apart
		# but we don't handle time zones so ignore that for now
		if {![string is integer $date]} {
			return false
		}
		return true
	}
}
