
package require struct::list

proc ::cnt::voting-types {} {
	set types [info class subclasses ::cnt::votings::voting]
	struct::list map $types {namespace tail}
}

proc ::cnt::voting {type args} {
	if {$type == "voting"} {
		throw {invalid voting} "Invalid voting type. Known types: [voting-types]"
	}
	set classname ::cnt::votings::$type
	try {
		info object class $classname oo::class
		set voting [$classname new {*}$args]
		$voting save
		return $voting
	} trap {TCL LOOKUP OBJECT} {} {
		throw {invalid voting} "Invalid voting type. Known types: [voting-types]"
	}
}

namespace eval ::cnt::votings {
	namespace path ::tcl::mathop

	# Generate a name for a plenary
	proc make-plenary-channel {} {
		set string [cnt::random-string 4 "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"]
		return "#plenary$string"
	}
}

oo::class create ::cnt::votings::voting {
	superclass ::cnt::db

	method construct {iline ithreshold {validate ""}} {
		namespace path [list {*}[namespace path] ::tcl::mathop]

		variable threshold $ithreshold
		variable line $iline
		variable options ""

		variable votes
		variable result
		variable created [clock seconds]

		my params

		if {$options == ""} {
			set options [list yes no]
		}
		lappend options abstain

		if {![info exists result]} {
			set result ""
		}

		eval [string map "%v [self object]" $validate]
	}

	method full-result {} {
		variable options
		variable votes
		array set results ""

		foreach option $options {
			set results($option) 0
		}
		foreach user [array names votes] {
			incr results($votes($user))
		}
		array get results
	}

	method vote {user vote {allowrepeat true}} {
		variable result
		variable options
		variable votes

		if {$result != ""} {
			throw {invalid vote finished} "This voting has already finished"
		}

		if {$user in [array names votes] && !$allowrepeat} {
			throw {invalid vote nochange} "Not allowed to change votes"
		}

		set possiblevotes [::struct::list filterfor option $options { [string first $vote [string tolower $option]] == 0}]

		switch [llength $possiblevotes] {
			0 {
				throw {invalid vote nooption} "There's no such option in this voting. Valid options are: [join $options]"
			}
			1 {
				set votes($user) $possiblevotes
				my update
			}
			default {
				throw {invalid vote ambiguity} "Too few letters. Can be any of: [join $possiblevotes]"
			}
		}
	}

	method plenary? {} {
		variable channel
		expr {$channel != [[my irc] mainchannel]}
	}

	method done? {voterscnt} {
		variable result

		if {$result != ""} {
			return true
		}

		array set results [my full-result]
		set threshold [my Get-threshold [- $voterscnt $results(abstain)]]
		unset results(abstain)

		if {$threshold < 0} {
			return false
		}

		foreach option [array names results] {
			if {$results($option) > $threshold} {
				set result $option
				my update
				return true
			}
		}
		return false
	}

	# groups - list of groups [power members...]
	# afk - what to do with the users who didn't vote.
	# 	lost - part of the group's power is lost
	# 	split - split the power among users who did vote
	method done-groups? {groups groupsize {afk lost}} {
		variable result
		variable votes
		variable options

		puts "GROUPS: $groups"
		puts "GROUPSIZE: $groupsize"

		if {$result != ""} {
			return true
		}

		set totalpower 0
		array set power [struct::list flatten -full [struct::list mapfor group $groups {
			if {[llength $group] == 1} continue
			set grouppower [lindex $group 0]
			incr totalpower $grouppower
			set groupusers [lrange $group 1 end]
			switch $afk {
				lost {
					set userpower [expr {$grouppower  / $groupsize}]
				}
				split {
					set userpower [expr {$grouppower / [llength $groupusers]}]
				}
			}
			struct::list mapfor user [lrange $group 1 end] {
				list $user $userpower
			}
		}]]
		puts "POWER: $power"
		puts "TOTALPOWER: $totalpower"

		array set results {}
		foreach option $options {
			set results($option) 0
		}

		foreach user [array names votes] {
			set vote $votes($user)
			incr results($vote) $power($user)
		}

		set threshold [expr {($totalpower - $results(abstain)) / 2.0}]

		foreach option [array names results] {
			if {$results($option) > $threshold} {
				set result $option
				my update
				return true
			}
		}
		return false
	}

	method timeout {voters defaultvote} {
		variable result

		if {$result == ""} return

		foreach voter $voters {
			vote $voter $defaultvote false
		}

		if {[done? [llength $voters]]} {
			return true
		} else {
			set result draw
			return false
		}
	}

	method Get-threshold {voterscnt} {
		variable threshold
		switch $threshold {
			majority	{/ $voterscnt 2}
			unanimous	{- $voterscnt 1}
			second		{+ 1}
			default		{
				throw {invalid threshold} "Invalid voting threshold set: $threshold"
			}
		}
	}

	method default-options? {} {
		variable options
		expr {[struct::set equal $options [list yes no abstain]] || $options == ""}
	}

	method get {var} {
		variable $var
		if {[info exists $var]} {
			set $var
		} else {
			error "Invalid variable name in object: $var"
		}
	}

	method tolist {} {
		variable line
		variable created
		variable threshold
		variable votes
		variable result
		variable channel

		list voting [my type] $line $created $threshold [array get votes] $result $channel
	}

	method fromlist {list} {
		variable line
		variable created
		variable threshold
		variable votes
		variable result
		variable channel

		lassign $list line created threshold votesdict result channel
		array set votes $votesdict
		my construct $line $threshold ;# No need to validate again
	}

	method result {} {
		variable result
		return $result
	}
}

proc defvoting {name body} {
	oo::class create ::cnt::votings::$name " \
		superclass ::cnt::votings::voting \
		$body"
}

proc split-topic {line} {
	set tokens [split $line ":"]
	set topic [lindex $tokens 0]
	set options [struct::list map [split [lindex $tokens 1] ","] {string trim}]
	list $topic $options
}

defvoting poll {
	method params {} {
		variable line

		variable topic
		variable options
		lassign [split-topic $line] topic options

		if {$topic == ""} {
			throw {invalid voting} "Topic can't be empty"
		}

		if {[llength $options] == 1} {
			throw {invalid voting} "Need at least two options"
		}
	}

	method summary {} {
		variable topic
		return $topic
	}
}

defvoting setting {
	method params {} {
		variable line
		variable topic $line
		variable setting [lindex $line 0]
		variable value [lrange $line 1 end]

		set options ""
		switch $setting {
			threshold {
				set options [list majority unanimous second]
			}
			default-vote {
				set options [list yes no abstain]
			}
			timeout {
				if {[llength $value] != 2 ||
					![string is integer [lindex $value 0]] ||
					[string trimright [lindex $value 1] s] ni [list second minute hour day week]
				} {
					throw {invalid voting} "Timeout should have the following form: <number> <unit>. E.g. 5 minutes or 1 week"
				}
				set value [my timeout-seconds $value]
			}
			security {
				if {$value ni [list none poll] &&
					[lindex $value 0] != "password"} {
					throw {invalid voting} "Allowed values for this setting: none, poll, password <password>"
				}
			}
			public-polls {
				set options [list yes no]
			}
			public-settings {
				set options [list yes no]
			}
			public-delegates {
				set options [list yes no]
			}
			default {
				throw {invalid voting} "Known settings: threshold, default-vote, timeout, security, public-polls, public-settings, public-delegates"
			}
		}
		if {$options != "" && $value ni $options} {
			throw {invalid voting} "Allowed values for this setting: [join $options \", \"]"
		}
	}

	method timeout-seconds {value} {
		set coeff [switch [lindex $value 1] {
			second	{expr 1}
			minute	{expr 60}
			hour	{expr 60 * 60}
			day		{expr 60 * 60 * 24}
			week	{expr 60 * 60 * 24 * 7}
		}]
		return [expr {[lindex $value 0] * $coeff}]
	}

	method summary {} {
		variable setting
		variable value
		return "Change setting $setting to $value"
	}
}

defvoting opcommand {
	method params {} {
		variable line
		variable topic $line
		variable command $line
	}

	method summary {} {
		variable command
		return "Execute: $command"
	}
}

defvoting invite {
	method params {} {
		variable line
		variable nick [lindex $line 0]
		variable reason [join [lrange $line 1 end]]
	}

	method summary {} {
		variable nick
		variable reason
		return "Invite $nick. Reason: $reason"
	}
}

defvoting send-delegate {
	method params {} {
		variable channel
		variable line

		variable otherchannel
		variable options
		variable delegate ""
		if {[string first : $line] != -1} {
			lassign [split-topic $line] otherchannel options
			variable users [dict create]
			struct::list mapfor nick $options {
				dict set users [hlirc::canonicalize $nick] [[my irc] nick-user $nick]
			}
		} else {
			lassign [split $line] delegate otherchannel
			variable users [dict create [hlirc::canonicalize $delegate] [[my irc] nick-user $delegate]]
		}

		if {$otherchannel == $channel} {
			throw {invalid voting} "You cannot delegate to and from the same channel"
		}
		if {![::hlirc::channel? $otherchannel]} {
			throw {invalid voting} {!start send-delegate <channel>: [delegate], [delegate]...}
		}

		if {$delegate == ""} {
			lappend options nobody
		}
	}

	method summary {} {
		variable otherchannel
		variable delegate
		if {$delegate == ""} {
			return "Send delegate to channel $otherchannel"
		} else {
			return "Send delegate $delegate to channel $otherchannel"
		}
	}

	method result {} {
		variable result
		variable delegate
		if {$result == "yes"} {
			return $delegate
		} else {
			return $result
		}
	}

	method result-user {} {
		variable result
		variable users
		if {$result != ""} {
			dict get $users [hlirc::canonicalize [my result]]
		} else {
			return ""
		}
	}
}

defvoting request-delegate {
	method params {} {
		variable line
		variable otherchannel $line
		if {![::hlirc::channel? $otherchannel] ||
	  		![my default-options?]} {
			throw {invalid voting} "!start request-delegate <channel>"
		}
	}

	method summary {} {
		variable otherchannel
		return "Request delegate from channel $otherchannel"
	}
}

defvoting recall-delegate {
	method params {} {
		variable line
		variable nick ""
		variable otherchannel ""
		foreach value [split $line] {
			if {[::hlirc::channel? $value]} {
				set otherchannel $value
			} elseif {[::hlirc::nick? $value]} {
				set nick $value
			} else {
				throw {invalid voting} "An argument is neither a channel nor a valid nick"
			}
		}
		if {![my default-options?]} {
			throw {invalid voting} {!start recall-delegate ([channel] [nick])}
		}
		if {$nick == "" && $otherchannel == ""} {
			throw {invalid voting} {!start recall-delegate ([channel] [nick])}
		}
		variable user [[my irc] nick-user $nick]
	}

	method summary {} {
		variable nick
		variable channel
		set nickstr $nick
		set channelstr $channel
		if {$nick == ""} {
			set nickstr "all delegates"
		}
		if {$channel == ""} {
			set channelstr "all channels"
		}
		return "Recall $nickstr from $channelstr"
	}
}

defvoting approve-delegate {
	method params {} {
		variable line
		set lst [split $line]
		variable otherchannel [lindex $lst 0]
		variable nick [lindex $lst 1]
	}

	method summary {} {
		variable nick
		variable otherchannel
		return "Approve delegate $nick from $otherchannel"
	}
}

defvoting plenary {
	method params {} {
		variable line
		variable channel
		set values [split $line]
		if {[hlirc::channel? [lindex $values 0]]} {
			set plenarychannel [lindex $values 0]
			struct::list shift values
		}
		if {[lindex $values 0] == "public"} {
			variable type public
			variable delegates 0
			variable channels [lrange $values 1 end]
		} elseif {[string is integer [lindex $values 0]] &&
				  [lindex $values 1] in [list delegate delegates]} {
			variable type delegates
			variable delegates [lindex $values 0]
			variable channels [lrange $values 2 end]
		} else {
			throw {invalid voting} {!start plenary [name] (public | <n> delegates) [channels...]}
		}

		if {![my default-options?]} {
			throw {invalid voting} {!start plenary [name] (public | <n> delegates) [channels...]}
		}

		struct::set include channels [hlirc::canonicalize $channel]

		variable plenarychannel [::cnt::votings::make-plenary-channel]
	}

	method summary {} {
		variable channels
		variable type
		variable delegates
		switch $type {
			public {
				set typestr public
			}
			delegates {
				if {$delegates == 1} {
					set typestr "1 delegate"
				} else {
					set typestr "$delegates delegates"
				}
			}
		}
		return "Assemble a $typestr plenary with [join $channels {, }]"
	}
}

defvoting attend-plenary {
	method params {} {
		variable line
		set tokens [split $line]
		variable plenarychannel [lindex $tokens 0]
		variable channels [lrange $tokens 1 end]
	}

	method summary {} {
		variable plenarychannel
		variable channels
		return "Attend a plenary on $plenarychannel with: [join $channels {, }]"
	}
}

# Plenary votings

defvoting invite-channel {
	method params {} {
		variable line
		variable channel-to-invite $line
	}

	method summary {} {
		variable channel-to-invite
		variable channel
		return "Invite channel ${channel-to-invite} to plenary $channel"
	}
}

defvoting kick-channel {
	method params {} {
		variable line
		variable channel-to-kick $line
	}

	method summary {} {
		variable channel-to-kick
		variable channel
		return "Invite channel ${channel-to-kick} to plenary $channel"
	}
}


